from __future__ import annotations

import io
import json
import os
import re
import types
from typing import Optional, Union

from bs4 import BeautifulSoup

import requests
from seeq.sdk import *
from seeq.spy import _common
from seeq.spy import _login
from seeq.spy._errors import *
from seeq.spy._redaction import safely, request_safely
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.workbooks import _item, _render
from seeq.spy.workbooks._content import Content, DateRange, AssetSelection
from seeq.spy.workbooks._item_map import ItemMap, OverrideItemMap
from seeq.spy.workbooks._workstep import AnalysisWorkstep

FIXUP_PREFIX = 'Annotation to fix up'


class Annotation:
    _html: str
    _images: dict

    def __init__(self, worksheet, annotation_type):
        self.annotation_type = annotation_type
        self.worksheet = worksheet
        self._html = ''
        self._images = dict()
        self.plots_to_render = set()
        if not _common.present(self.worksheet, 'Annotation ID'):
            self.worksheet['Annotation ID'] = _common.new_placeholder_guid()

    @property
    def id(self):
        # We store this on the worksheet because it's the only piece of metadata we need for the annotation and it's
        # convenient to have it in the worksheet's JSON when saved.
        return _common.get(self.worksheet, 'Annotation ID')

    @id.setter
    def id(self, _id):
        self.worksheet['Annotation ID'] = _id

    def refresh_from(self, new_item, item_map: ItemMap, status: Status):
        self.annotation_type = new_item.annotation_type
        # Note that we purposefully don't touch the worksheet reference, since it will have stayed the same
        self._set_html(new_item.html)
        self._images = new_item.images

    def _find_image_references(self):
        if not self.html:
            return list()

        matches = re.finditer(r'src="/api(/annotations/(.*?)/images/(.*?))"', self.html)
        return [(match.group(1), match.group(2), match.group(3)) for match in matches]

    @property
    def html(self):
        return self._html

    @html.setter
    def html(self, value):
        self._set_html(value)

    def _set_html(self, value):
        if value is None:
            self._html = ''
        else:
            self._html = value

    @property
    def images(self):
        return self._images

    @property
    def referenced_items(self):
        return list()

    @property
    def referenced_worksteps(self):
        return self._find_workstep_references()

    def _find_workstep_references(self):
        return set()

    def find_workbook_links(self, session: Session, status: Status):
        if not self.html:
            return dict()

        url = _common.get(self.worksheet.workbook, 'Original Server URL')
        if not url:
            return dict()

        # TODO can this be converted to use the _common.workbook_worksheet_url_regex methods?
        edit_link_no_folder_regex = \
            r'%s/workbook/(?P<workbook>%s)/worksheet/(?P<worksheet>%s)' % (url,
                                                                           _common.GUID_REGEX,
                                                                           _common.GUID_REGEX)

        edit_link_with_folder_regex = \
            r'%s/%s/workbook/(?P<workbook>%s)/worksheet/(?P<worksheet>%s)' % (url,
                                                                              _common.GUID_REGEX,
                                                                              _common.GUID_REGEX,
                                                                              _common.GUID_REGEX)

        view_link_regex = \
            r'%s/view/(?P<worksheet>%s)' % (url, _common.GUID_REGEX)

        present_link_regex = \
            r'%s/present/worksheet/(?P<workbook>%s)/(?P<worksheet>%s)' % (url,
                                                                          _common.GUID_REGEX,
                                                                          _common.GUID_REGEX)

        workstep_tuples = dict()
        for regex in [edit_link_no_folder_regex, edit_link_with_folder_regex, view_link_regex, present_link_regex]:
            matches = re.finditer(regex, self.html, re.IGNORECASE)

            for match in matches:
                group_dict = dict(match.groupdict())
                if 'workbook' not in group_dict:
                    items_api = ItemsApi(session.client)
                    _worksheet_id = group_dict['worksheet']
                    item_output = safely(
                        lambda: items_api.get_item_and_all_properties(id=_worksheet_id),
                        action_description=f'get the Workbook ID for Worksheet {_worksheet_id}',
                        status=status)  # type: ItemOutputV1
                    if item_output is not None:
                        group_dict['workbook'] = item_output.workbook_id

                if group_dict['workbook'].upper() not in workstep_tuples:
                    workstep_tuples[group_dict['workbook'].upper()] = set()

                workstep_tuples[group_dict['workbook'].upper()].add(
                    (group_dict['workbook'].upper(), group_dict['worksheet'].upper(), None))

        return workstep_tuples

    def pull(self, session: Session, *, include_images=True, status=None):
        session = Session.validate(session)
        self._images = dict()
        annotations_api = AnnotationsApi(session.client)
        # Note: get_annotations() won't error if the user doesn't have access to the `annotates` IDs
        _annotations = annotations_api.get_annotations(annotates=[self.worksheet.id])  # type: AnnotationListOutputV1

        annotation_output = None
        for annotation_item in _annotations.items:  # type: AnnotationOutputV1
            candidate = safely(lambda: annotations_api.get_annotation(id=annotation_item.id),
                               action_description=f'get details for Annotation {annotation_item.id}',
                               status=status)  # AnnotationOutputV1
            if candidate is not None and candidate.type == self.annotation_type:
                annotation_output = candidate
                break

        if not annotation_output:
            # This seems possible (although rare) and it seems like the annotation gets created immediately if the
            # user clicks on the Journal tab. So just return nothing but allow this Annotation object to persist with
            # a placeholder ID.
            return None

        self.id = annotation_output.id
        self._set_html(annotation_output.document)

        if include_images:
            image_references = self._find_image_references()
            for query_params, annotation_id, image_id in image_references:
                if (annotation_id, image_id) in self.images:
                    continue

                self.worksheet.workbook.update_status('Pulling image', 1)

                api_client_url = session.get_api_url()
                request_url = api_client_url + query_params

                self.images[(annotation_id, image_id)] = _login.pull_image(session, request_url)

        return annotation_output

    def push(self, session: Session, pushed_workbook_id, pushed_worksheet_id, item_map, datasource_output,
             push_images, label, status=None):
        self.render_plots()

        new_annotation = AnnotationInputV1()
        new_annotation.document = 'Placeholder document'
        new_annotation.name = 'Unnamed'
        new_annotation.description = 'Placeholder description'
        new_annotation.type = self.annotation_type

        annotations_api = AnnotationsApi(session.client)
        # Note: get_annotations() won't error if the user doesn't have access to the `annotates` IDs
        existing_annotations = annotations_api.get_annotations(
            annotates=[pushed_worksheet_id])  # type: AnnotationListOutputV1

        relevant_annotations = [a for a in existing_annotations.items if a.type == self.annotation_type]
        if len(relevant_annotations) == 0:
            if isinstance(self, Report):
                # Creating a report requires an OptionalReportInputV1
                new_annotation.report_input = OptionalReportInputV1()
            else:
                new_annotation.interests = [
                    AnnotationInterestInputV1(interest_id=pushed_worksheet_id)
                ]

                if isinstance(self, Journal):
                    # Reports cannot have an interest to the workbook, see CRAB-18738
                    new_annotation.interests.append(AnnotationInterestInputV1(interest_id=pushed_workbook_id))

            relevant_annotation = safely(lambda: annotations_api.create_annotation(body=new_annotation),
                                         action_description=f'create {new_annotation.type} {new_annotation.name}',
                                         status=status)  # type: AnnotationOutputV1
            if relevant_annotation is None:
                return
        else:
            relevant_annotation = relevant_annotations[0]

        item_map[self.id] = relevant_annotation.id

        html = self._push_specific(session, item_map, datasource_output, label, new_annotation, relevant_annotation)

        # Create an override map for image mapping, which can be different for the same annotation in the case of a
        # template
        images_map = dict()

        # First add a prefix to existing image references in the HTML so that we can handle cases where
        # the names that are picked by "POST /annotations/{id}/images" overlap with existing names and would
        # otherwise cause the find & replace that happens later to be non-deterministic
        image_prefix = 'ToBeMapped_'
        html = re.sub(r'src="(/api/annotations/.*?/images/)(.*?)"', rf'src="\1{image_prefix}\2"', html)

        if push_images:
            item_map[f'Images for {self.id}'] = images_map

            for _, annotation_id, image_id in self._find_image_references():
                api_client_url = session.get_api_url()
                request_url = api_client_url + '/annotations/%s/images' % relevant_annotation.id

                self.worksheet.workbook.update_status('Pushing image', 1)

                response = requests.post(url=request_url,
                                         files={
                                             "file": (image_id, io.BytesIO(self.images[(annotation_id, image_id)]))
                                         },
                                         headers={
                                             "Accept": "application/vnd.seeq.v1+json",
                                             "x-sq-auth": session.client.auth_token
                                         },
                                         verify=session.https_verify_ssl)

                if response.status_code != 201:
                    raise SPyRuntimeError(
                        f'Could not upload image file {image_id} for worksheet {pushed_worksheet_id}:\n'
                        f'Response code: {response.status_code}\n'
                        f'Response content: {response.content}')

                link_json = json.loads(response.content)

                match = re.match(r'.*?images/(.*)', link_json['link'])
                new_image_id = match.group(1)

                images_map[f'{image_prefix}{image_id}'] = new_image_id
        else:
            images_map = item_map[f'Images for {self.id}']

        bs = BeautifulSoup(html, features='html.parser')
        find_result = bs.find_all(['p', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'title'])
        name = 'Unnamed'
        description = None

        if len(find_result) > 0:
            name = ' '.join(re.split(r'[\s\n]+', find_result[0].get_text().strip())[:20])
        if len(find_result) > 1:
            description = ' '.join(re.split(r'[\s\n]+', find_result[1].get_text().strip())[:50])

        html = self._replace_items_in_html(OverrideItemMap(item_map, override_map=images_map), html)

        worksheet_link_replacement = r'links?type=workstep&amp;workbook=%s&amp;worksheet=%s&amp;' % (
            pushed_workbook_id, pushed_worksheet_id
        )

        html = re.sub(_common.WORKSHEET_LINK_REGEX, worksheet_link_replacement, html, flags=re.IGNORECASE)

        original_server_url = _common.get(self.worksheet.workbook, 'Original Server URL')
        new_server_url = _item.get_canonical_server_url(session)
        if original_server_url is not None and new_server_url is not None:
            item_map[original_server_url] = new_server_url

        new_annotation.name = name if len(name.strip()) > 0 else 'Unnamed'
        new_annotation.description = description
        new_annotation.document = html
        new_annotation.interests = list()
        for interest in relevant_annotation.interests:  # type: AnnotationInterestOutputV1
            interest_item = interest.item  # type: ItemPreviewV1
            # At Chevron, we encountered a case where there were multiple interests returned with the same ID, which
            # caused Appserver to choke when updating the annotation. So filter those out.
            if any(interest_item.id == i.interest_id for i in new_annotation.interests):
                continue
            if isinstance(self, Report) and interest_item.id == pushed_workbook_id:
                # Reports cannot have an interest to the workbook, see CRAB-18738
                continue
            new_interest = AnnotationInterestInputV1()
            new_interest.interest_id = interest_item.id
            new_interest.detail_id = interest.capsule
            new_annotation.interests.append(new_interest)

        new_annotation.created_by_id = relevant_annotation.created_by.id
        safely(lambda: annotations_api.update_annotation(id=relevant_annotation.id, body=new_annotation),
               action_description=f'update Annotation {relevant_annotation.id}',
               status=status)

        item_map[f'{FIXUP_PREFIX} - {self.id}'] = types.SimpleNamespace(id=relevant_annotation.id, body=new_annotation)

    @staticmethod
    def push_fixups(session: Session, status: Status, item_map: ItemMap):
        annotations_api = AnnotationsApi(session.client)

        for key in item_map.keys():
            if not key.startswith(FIXUP_PREFIX):
                continue

            annotation = item_map[key]

            annotation.body.document = _item.replace_items(annotation.body.document, item_map)

            safely(lambda: annotations_api.update_annotation(id=annotation.id, body=annotation.body),
                   action_description=f'fixup Annotation {annotation.id}',
                   status=status)

    def _replace_items_in_html(self, item_map: ItemMap, html):
        # When a workbook is duplicated via the Workbench UI, the workstep links within Journals actually refer to
        # the original workbook. This works in the UI because workstep content has no real dependency on the
        # workbook/worksheet they're associated with. When pulling, we accommodate this by pulling a Workstep and
        # associating it with the "proper" Worksheet object, but then during push we have to fix up the links in case
        # the "original" workbook/worksheet wasn't included in the workbooks to be pushed.
        workstep_map = _common.get(item_map, self.worksheet.item_map_worksteps_key())
        if workstep_map:
            html = _item.replace_items(html, workstep_map)

        html = _item.replace_items(html, item_map)

        return html

    def _push_specific(self, session: Session, item_map, datasource_output, label, new_annotation, annotation,
                       override_content_dict=None, status=None):
        # This will be overridden in derived classes to do some specific work like pushing Content and DateRanges
        return self.html

    def _get_annotation_file(self, workbook_folder):
        return os.path.join(workbook_folder, '%s_%s' % (self.annotation_type, self.worksheet.id))

    @staticmethod
    def _get_image_file(workbook_folder, image_id_tuple):
        return os.path.join(workbook_folder, 'Image_%s_%s' % image_id_tuple)

    @staticmethod
    def _get_html_attr(fragment, attribute):
        attr_match = re.findall(r'\s+%s="(.*?)"' % attribute, fragment)
        return attr_match[0] if len(attr_match) > 0 else None

    def save(self, workbook_folder, *, include_rendered_content=False, pretty_print_html=False):
        self.render_plots()
        html_file = f'{self._get_annotation_file(workbook_folder)}.html'
        with open(html_file, 'w', encoding='utf-8') as f:
            if self.html:
                if pretty_print_html:
                    html_to_save = BeautifulSoup(self.html, features='html.parser').prettify()
                    html_to_save = _common.fix_up_ckeditor_curly_brace_weirdness(html_to_save)
                    # If we don't trim the spaces within <a> tags, you'll get extra spaces underlined in the UI
                    html_to_save = re.sub(r'(<a .*?>)[\s\n]+(.*?)[\s\n]+(</a>)', r'\1\2\3', html_to_save)
                else:
                    html_to_save = self.html

                f.write(html_to_save)

        _common.save_image_files(self.images, workbook_folder)

    def _load(self, workbook_folder):
        html_file = f'{self._get_annotation_file(workbook_folder)}.html'

        with open(html_file, 'r', encoding='utf-8') as f:
            self.html = f.read()

        matches = re.finditer(r'src="/api(/annotations/(.*?)/images/(.*?))"', self.html)
        for match in matches:
            image_id_tuple = (match.group(2), match.group(3))
            image_file = _common.get_image_file(workbook_folder, image_id_tuple)

            with open(image_file, 'rb') as f:
                self.images[image_id_tuple] = f.read()

    def add_image(self, *, filename=None, buffer=None, image_format=None, placement=None, just_src=False):
        """
        Add an image to the annotation.

        Parameters
        ----------
        filename: str
            The full path to the image file
        buffer: str
            The bytes of the image in memory (must also specify image_format)
        image_format
            The image format of what is supplied in bytes (e.g. 'png', 'jpg')
        placement : {'end', 'beginning', None}, default None
            The location to add the image to an existing document.
        just_src : bool
            False if full <img> html tags desired, True if you just want the
            url to put in the <img src="<url>"> attribute yourself.
        """
        if filename and buffer:
            raise SPyValueError('Either filename or buffer must be supplied to image function -- not both')

        if buffer and not image_format:
            raise SPyValueError('image_format must be specified if buffer is supplied')

        if placement not in ['end', 'beginning', None]:
            raise SPyValueError(f"placement must be one of {['end', 'beginning', None]}")

        if placement and just_src:
            raise SPyValueError(f"placement must None if just_src is True")

        html = self.html
        if filename:
            image_name = os.path.basename(filename)
            with open(filename, 'rb') as img:
                self.images[(self.id, image_name)] = img.read()
        else:
            image_name = f'{_common.new_placeholder_guid()}.{image_format}'
            self.images[(self.id, image_name)] = buffer

        url = f'/api/annotations/{self.id}/images/{image_name}'
        if just_src:
            return url

        image_html = f'<img class="fr-fic fr-fin fr-dii" src="{url}">'
        if placement is not None:
            if placement == 'beginning':
                html = image_html + html
            else:
                html += image_html
            self._set_html(html)

        return image_html

    def add_plot_to_render(self, plot_render_info, date_range):
        image_id = _common.new_placeholder_guid()
        filename = f'{image_id}.{plot_render_info.image_format}'
        self.plots_to_render.add((self.id, filename, plot_render_info, date_range))
        return f'<img class="fr-fic fr-fin fr-dii" src="/api/annotations/{self.id}/images/{filename}"/>'

    def render_plots(self):
        for annotation_id, filename, plot_render_info, date_range in self.plots_to_render:
            self.images[annotation_id, filename] = plot_render_info.render_function(date_range)

        self.plots_to_render.clear()


class Journal(Annotation):

    def __init__(self, worksheet):
        super().__init__(worksheet, 'Journal')

    @staticmethod
    def load(worksheet, workbook_folder):
        journal = Journal(worksheet)
        journal._load(workbook_folder)
        return journal

    @property
    def referenced_items(self):
        referenced_items = list()
        if self.html:
            matches = re.finditer(r'item%s(%s)' % (_common.HTML_EQUALS_REGEX, _common.GUID_REGEX), self.html,
                                  re.IGNORECASE)
            for match in matches:
                referenced_items.append(_item.Reference(match.group(1).upper(), _item.Reference.JOURNAL,
                                                        self.worksheet))

        return referenced_items

    def _find_workstep_references(self):
        if not self.html:
            return set()

        workstep_references = set()
        regex = r'workbook%s(%s)&amp;worksheet%s(%s)&amp;workstep%s(%s)' % (
            _common.HTML_EQUALS_REGEX, _common.GUID_REGEX,
            _common.HTML_EQUALS_REGEX, _common.GUID_REGEX,
            _common.HTML_EQUALS_REGEX, _common.GUID_REGEX)
        matches = re.finditer(regex, self.html, re.IGNORECASE)

        for match in matches:
            workstep_references.add((match.group(1).upper(), match.group(2).upper(), match.group(3).upper()))

        return workstep_references


class Report(Annotation):
    rendered_content_images: Optional[dict]

    def __init__(self, worksheet):
        super().__init__(worksheet, 'Report')
        self.rendered_content_images = None

        self.date_ranges = dict()
        self.asset_selections = dict()
        self.content = dict()
        self.schedule = None
        self.data = None

    @property
    def referenced_items(self):
        referenced_items = list()

        for date_range in self.date_ranges.values():
            if _common.present(date_range, 'Condition ID'):
                referenced_items.append(
                    _item.Reference(date_range['Condition ID'], _item.Reference.DATE_RANGE_CONDITION,
                                    self.worksheet))

        for asset_selection in self.asset_selections.values():
            if _common.present(asset_selection, 'Asset ID'):
                referenced_items.append(_item.Reference(asset_selection['Asset ID'], _item.Reference.ASSET_SELECTION,
                                                        self.worksheet))

        return referenced_items

    def pull(self, session: Session, include_images=True, status=None):
        session = Session.validate(session)
        annotation_output = super().pull(session, include_images=include_images, status=status)

        if annotation_output is None:
            return

        content_api = ContentApi(session.client)
        contents = safely(
            lambda: content_api.get_contents_with_all_metadata(report_id=annotation_output.id),
            action_description=f'get Content items within Report {annotation_output.id}',
            status=status)
        if contents is None:
            return

        for content in contents.content_items:
            new_date_range = None
            if content.date_range is not None:
                new_date_range = DateRange.pull(content.date_range, report=self, session=session, status=status)
                self.date_ranges[new_date_range.id] = new_date_range

            new_asset_selection = None
            if content.asset_selection is not None:
                new_asset_selection = AssetSelection.pull(content.asset_selection, report=self, session=session,
                                                          status=status)
                self.asset_selections[new_asset_selection.id] = new_asset_selection

            new_content_definition = {
                'Name': content.name,
                'ID': content.id,
                'Width': content.width,
                'Height': content.height,
                'Workbook ID': content.source_workbook,
                'Worksheet ID': content.source_worksheet,
                'Workstep ID': content.source_workstep,
                'Scale': content.scale,
                'selector': content.selector,
                'Summary Type': content.summary_type,
                'Summary Value': content.summary_value,
                'Date Range ID': content.date_range.id if new_date_range is not None else None,
                'Asset Selection ID': content.asset_selection.id if new_asset_selection is not None else None,
            }

            new_content = Content(new_content_definition, self)
            self.content[new_content.definition['ID']] = new_content

        # Go back through the date range IDs on the annotation in case they aren't associated with any content and
        # therefore won't be in the output from the get_contents_with_all_metadata call.
        for date_range_id in annotation_output.date_range_ids:
            if date_range_id not in self.date_ranges.keys():
                new_date_range = DateRange.pull(date_range_id, report=self, session=session, status=status)
                self.date_ranges[new_date_range.id] = new_date_range

        # Go back through the asset selection IDs on the annotation in case they aren't associated with any content
        # and therefore won't be in the output from the get_contents_with_all_metadata call.
        for asset_selection_id in annotation_output.asset_selection_ids:
            if asset_selection_id not in self.asset_selections.keys():
                new_asset_selection = AssetSelection.pull(asset_selection_id, report=self, session=session,
                                                          status=status)
                self.asset_selections[new_asset_selection.id] = new_asset_selection

        # There is a cron schedule only if it is not null or the list is not empty
        if annotation_output.cron_schedule:
            self.schedule = {
                'Cron Schedule': annotation_output.cron_schedule,
                'Background': annotation_output.background
            }

    @staticmethod
    def _refresh_child_dict(existing_item_dict: dict, new_item_dict: dict, item_map: ItemMap, status: Status):
        items = list(existing_item_dict.items())
        for _original_id, existing_item in items:
            if existing_item.id not in item_map or item_map[existing_item.id] not in new_item_dict:
                # The Content, AssetSelection, or DateRange failed to get pushed. Keep going.
                existing_item_dict.pop(existing_item.id, None)
                continue

            new_item_id = item_map[existing_item.id]
            new_item = new_item_dict[new_item_id]

            # Update the map
            del existing_item_dict[_original_id]
            existing_item_dict[new_item_id] = existing_item

            # Refresh item itself
            existing_item.refresh_from(new_item, item_map, status)

    def refresh_from(self, new_item, item_map: ItemMap, status: Status):
        super().refresh_from(new_item, item_map, status)

        Report._refresh_child_dict(self.date_ranges, new_item.date_ranges, item_map, status)
        Report._refresh_child_dict(self.asset_selections, new_item.asset_selections, item_map, status)
        Report._refresh_child_dict(self.content, new_item.content, item_map, status)

    def _push_specific(self, session: Session, item_map: ItemMap, datasource_output, label, new_annotation,
                       existing_annotation, override_content_dict=None, status=None):
        date_range_ids_to_archive = list(existing_annotation.date_range_ids)
        asset_selection_ids_to_archive = list(existing_annotation.asset_selection_ids)
        content_ids_to_archive = list(existing_annotation.content_ids)

        existing_date_ranges = dict()
        for date_range_id in existing_annotation.date_range_ids:
            existing_date_range = DateRange.pull(date_range_id, session=session, status=status)
            existing_date_ranges[existing_date_range.id] = existing_date_range
            existing_date_ranges[existing_date_range.name] = existing_date_range

        existing_asset_selections = dict()
        for asset_selection_id in existing_annotation.asset_selection_ids:
            existing_asset_selection = AssetSelection.pull(asset_selection_id, session=session, status=status)
            existing_asset_selections[existing_asset_selection.id] = existing_asset_selection
            existing_asset_selections[existing_asset_selection.name] = existing_asset_selection

        existing_contents = dict()
        for content_id in existing_annotation.content_ids:
            existing_content = Content.pull(content_id, session=session, status=status)
            existing_contents[existing_content.id] = existing_content
            existing_contents[existing_content.name] = existing_content

        # Pushing date ranges, asset selections, and content has to be done in a couple of steps. First, we push the
        # date ranges and get their corresponding Seeq guids. Next, we push the asset selections and get their
        # corresponding Seeq guids. Then, we update the content to use those date range and asset selection guids so
        # that we can create content in the backend with the correct date ranges and asset selections. Finally,
        # we push up the report. The association to a report is made by identifying the reportId when creating or
        # updating the date ranges, asset selections, and/or content. The existing content, date ranges,
        # and asset selections here already have the correct reportIds attached.
        optional_report = OptionalReportInputV1(enabled=False, cron_schedule=None)

        def _push_it(_item_type, _item_dict, _existing, _ids_to_archive):
            for _item_object in _item_dict.values():  # type: Union[DateRange, AssetSelection, Content]
                try:
                    # This push has custom error handling so always use errors='raise'
                    _item_output = _item_object.push(session, item_map, _existing, status=Status(errors='raise'))
                    if _item_output.id in _ids_to_archive:
                        _ids_to_archive.remove(_item_output.id)
                except ApiException as e:
                    self.worksheet.workbook.status.raise_or_catalog(
                        f'Error processing {_item_type}: {_item_object}\n{_common.format_exception(e)}')

        _push_it('Date Range', self.date_ranges, existing_date_ranges, date_range_ids_to_archive)
        _push_it('Asset Selection', self.asset_selections, existing_asset_selections, asset_selection_ids_to_archive)
        _push_it('Content', self.content if override_content_dict is None else override_content_dict,
                 existing_contents, content_ids_to_archive)

        if self.schedule is not None:
            optional_report.enabled = True
            optional_report.background = self.schedule['Background']
            optional_report.cron_schedule = self.schedule['Cron Schedule']

        content_api = ContentApi(session.client)
        items_api = ItemsApi(session.client)
        for content_id_to_archive in content_ids_to_archive:
            @request_safely(action_description=f'archive Content {content_id_to_archive}', status=status)
            def _archive_content():
                # Something about this call is importantly different from the regular archive_item(). Using that
                #  method makes the Content creation act like POST instead of like PUT calls for subsequent runs.
                content_output = content_api.get_content(id=content_id_to_archive)  # type: ContentOutputV1
                content_api.update_content(id=content_id_to_archive,
                                           body=ContentInputV1(name='SPy archived this',
                                                               date_range_id=None,
                                                               asset_selection_id=None,
                                                               width=content_output.width,
                                                               height=content_output.height,
                                                               worksheet_id=content_output.source_worksheet,
                                                               workstep_id=content_output.source_workstep,
                                                               selector=content_output.selector,
                                                               summary_type=content_output.summary_type,
                                                               summary_value=content_output.summary_value,
                                                               report_id=None,
                                                               archived=True))

            _archive_content()

        for asset_selection_id_to_archive in asset_selection_ids_to_archive:
            safely(
                lambda: items_api.archive_item(id=asset_selection_id_to_archive),
                action_description=f'archive Asset Selection {asset_selection_id_to_archive}',
                status=status
            )

        for date_range_id_to_archive in date_range_ids_to_archive:
            safely(
                lambda: items_api.archive_item(id=date_range_id_to_archive),
                action_description=f'archive Date Range {date_range_id_to_archive}',
                status=status
            )

        new_annotation.report_input = optional_report

        return self.html

    def save(self, workbook_folder, *, include_rendered_content=False, pretty_print_html=False):
        super().save(workbook_folder, include_rendered_content=include_rendered_content,
                     pretty_print_html=pretty_print_html)

        json_file = f'{self._get_annotation_file(workbook_folder)}.json'
        json_dict = {
            'Schedule': self.schedule,
            'Date Ranges': [d.definition_dict for d in self.date_ranges.values()],
            'Asset Selections': [a.definition_dict for a in self.asset_selections.values()],
            'Content': [c.definition_dict for c in self.content.values()]
        }
        with open(json_file, 'w', encoding='utf-8') as f:
            json.dump(json_dict, f, indent=4, sort_keys=True)

        if include_rendered_content:
            _render.save(self, workbook_folder)

    @staticmethod
    def load(worksheet, workbook_folder):
        report = Report(worksheet)
        report._load(workbook_folder)
        return report

    def _load(self, workbook_folder):
        super()._load(workbook_folder)

        json_file = f'{self._get_annotation_file(workbook_folder)}.json'
        with open(json_file, 'r', encoding='utf-8') as f:
            json_dict = json.load(f)

        self.schedule = json_dict['Schedule']

        for date_range_dict in json_dict['Date Ranges']:
            new_date_range = DateRange(date_range_dict, self)
            self.date_ranges[new_date_range.id] = new_date_range

        if 'Asset Selections' in json_dict:
            for asset_selection_dict in json_dict['Asset Selections']:
                new_asset_selection = AssetSelection(asset_selection_dict, self)
                self.asset_selections[new_asset_selection.id] = new_asset_selection

        for content_dict in json_dict['Content']:
            new_content = Content(content_dict, self)
            self.content[new_content.id] = new_content

    def pull_rendered_content(self, session: Session, status: Status):
        _render.pull(session, self, status)

    def get_embedded_content_html(self,
                                  display,  # type: AnalysisWorkstep
                                  date_range=None,  # type: Optional[DateRange]
                                  size='medium',  # type: str
                                  shape='rectangle',  # type: str
                                  width=None,  # type: Optional[int]
                                  height=None,  # type: Optional[int]
                                  scale=1.0,  # type: float
                                  selector='',  # type: str
                                  asset_selection=None,  # type: Optional[AssetSelection]
                                  summary_type=None,  # type: Optional[str]
                                  summary_value=None  # type: Optional[str]
                                  ):
        # type: (...) -> str
        if width is None and height is None:
            shape = Content.CONTENT_SHAPE[shape]
            width = Content.CONTENT_SIZE[size]
            height = int(width * shape['height'] / shape['width'])
        elif width is None and height is not None:
            raise SPyValueError('You must specify a width if you specify a height')
        elif height is None and width is not None:
            raise SPyValueError('You must specify a height if you specify a width')

        new_content_definition = {'Name': f'SPy_content_{_common.new_placeholder_guid()}',
                                  'Width': width,
                                  'Height': height,
                                  'Workbook ID': display.worksheet.workbook.id,
                                  'Worksheet ID': display.worksheet.id,
                                  'Workstep ID': display.id,
                                  'selector': selector,
                                  'Scale': float(scale)}

        new_content = Content(new_content_definition, self)
        self.content[new_content.id] = new_content

        if date_range is not None:
            existing_date_range = None
            # If a date range with an existing name is input, use that date range for the new content
            for dr in self.date_ranges.values():
                if dr.definition['Name'] == date_range.definition['Name']:
                    existing_date_range = dr

            if existing_date_range is not None:
                new_content.definition['Date Range ID'] = existing_date_range.id
            else:
                new_date_range = DateRange(date_range.definition, self)
                new_content = Content(new_content_definition, self)
                new_content.definition['Date Range ID'] = new_date_range.id
                self.date_ranges[new_date_range.id] = new_date_range

        if asset_selection is not None:
            existing_asset_selection = None
            # If an asset selection with an existing name is input, use that asset selection for the new content
            for selection in self.asset_selections.values():
                if selection.definition['Name'] == asset_selection.definition['Name']:
                    existing_asset_selection = selection

            if existing_asset_selection is not None:
                new_content.definition['Asset Selection ID'] = existing_asset_selection.id
            else:
                new_asset_selection = AssetSelection(asset_selection.definition, self)
                new_content = Content(new_content_definition, self)
                new_content.definition['Asset Selection ID'] = new_asset_selection.id
                self.asset_selections[new_asset_selection.id] = new_asset_selection

        if summary_type is not None:
            new_content.definition['Summary Type'] = summary_type
        if summary_value is not None:
            new_content.definition['Summary Value'] = summary_value

        return new_content.html

    def _find_workstep_references(self):
        if not self.html:
            return set()

        workstep_references = set()
        for content in self.content.values():
            workstep_references.add((content.definition['Workbook ID'], content.definition['Worksheet ID'],
                                     content.definition['Workstep ID']))

        return workstep_references



from __future__ import annotations

import logging
import os
import sys
from pathlib import Path
from typing import Union

import nbconvert
import nbformat
from nbformat import NotebookNode

from seeq.spy import _datalab
from seeq.spy._errors import *

RESULTS_FOLDER = '_Job Results/'
CELL_EXECUTION_TIMEOUT = 86400


class ExecutionInstance:
    _logger = None

    def __init__(self):
        if not _datalab.is_executor():
            raise SPyRuntimeError('Execution of notebooks is not supported through SPy interface.')

        self.file_path = os.environ.get('SEEQ_SDL_FILE_PATH', '')
        self.label = os.environ.get('SEEQ_SDL_LABEL', '')
        self.index = os.environ.get('SEEQ_SDL_SCHEDULE_INDEX', '')
        self.job_key = os.environ.get('SEEQ_SDL_JOB_KEY', '')
        scheduled_file_filename: str = Path(self.file_path).name
        scheduled_file_folder: Path = Path(self.file_path).parent
        # merged_notebook_path is our new scheduled notebook with login cell
        self.merged_notebook_path = _compose_filename(self.index, self.label,
                                                      scheduled_file_filename,
                                                      scheduled_file_folder,
                                                      extension='.ipynb', result=False)
        # This is our resulting html file after execution
        self.job_result_path = _compose_filename(self.index, self.label,
                                                 scheduled_file_filename,
                                                 scheduled_file_folder,
                                                 extension='.html', result=True)

    @property
    def logger(self) -> logging.Logger:
        """
        Logger to be used for logging inside executor container
        """
        if self._logger is not None:
            return self._logger

        log_level = get_log_level_from_executor()

        # python logging doesnt have TRACE
        log_level = "DEBUG" if log_level == "TRACE" else log_level

        executor_logger = logging.getLogger("executor_logger")
        exec_handler = logging.StreamHandler(sys.stdout)
        exec_formatter = logging.Formatter(
            f'%(levelname)s - Notebook {self.file_path.replace("%", "%%")} with jobKey {self.job_key} %(message)s')
        exec_handler.setFormatter(exec_formatter)
        executor_logger.addHandler(exec_handler)
        executor_logger.setLevel(log_level)

        self._logger = executor_logger
        return executor_logger

    def execute(self):
        # Abort if scheduled notebook does not exist
        if not Path(self.file_path).exists():
            # TODO CRAB-25884: unschedule the notebook if no longer exists and notify the creator
            self.logger.error('not found, aborting')
            return

        # noinspection PyBroadException
        try:
            # Create the new scheduled notebook with login cell
            self.create_merged_notebook()

            # Execute the new scheduled notebook with login cell
            self.execute_merged_notebook()

            # Convert to html and save
            self.export_html()

            # Log success for executor
            self.logger.info('succeeded')
        except BaseException as e:
            self.logger.error('encountered error', exc_info=e)
        finally:
            # Cleanup
            if Path(self.merged_notebook_path).exists():
                Path(self.merged_notebook_path).unlink()

    def create_merged_notebook(self):
        # Open the notebook that has been scheduled for execution
        with open(self.file_path) as f_notebook_scheduled:
            nb_notebook_scheduled = nbformat.read(f_notebook_scheduled, nbformat.NO_CONVERT)

        # Get kernel language
        language = _datalab.get_notebook_language(nb_notebook_scheduled)
        if not language:
            error_message = f'could not determine language for {f_notebook_scheduled}'
            self.logger.error(error_message)
            raise SPyRuntimeError(error_message)

        # Get execution notebook
        execution_notebook = _datalab.get_execution_notebook(language)
        if not execution_notebook:
            error_message = f'could not find execution notebook for {f_notebook_scheduled} with language {language}'
            self.logger.error(error_message)
            raise SPyRuntimeError(error_message)

        # Open the dummy notebook with spy.login cell
        with open(execution_notebook) as f_notebook_execution:
            nb_notebook_execution = nbformat.read(f_notebook_execution, nbformat.NO_CONVERT)

        # Create new notebook dynamically that includes login cell first
        nb_notebook_merged = NotebookNode(nb_notebook_execution.copy())

        # Add in cells from scheduled notebook
        nb_notebook_merged['cells'].extend(nb_notebook_scheduled.cells.copy())

        # Write out the new joined notebook as hidden notebook with the same name as scheduled notebook
        with open(self.merged_notebook_path, 'w') as f_notebook_merged:
            nbformat.write(nb_notebook_merged, f_notebook_merged)

        # Log to executor
        self.logger.debug('successfully merged execution notebook with scheduled notebook')

    def execute_merged_notebook(self) -> NotebookNode:
        # Open the notebook for execution
        with open(self.merged_notebook_path, 'r+') as f_notebook_merged:
            nb_notebook_merged = nbformat.read(f_notebook_merged, nbformat.NO_CONVERT)

        # Configure the execute processor to allow errors and the output path
        proc = nbconvert.preprocessors.ExecutePreprocessor(timeout=CELL_EXECUTION_TIMEOUT,
                                                           allow_errors=True)
        proc.preprocess(nb_notebook_merged, {'metadata': {'path': Path(self.file_path).parent}})

        # Log to executor
        self.logger.debug('successfully executed merged notebook')

        # Python logger has no TRACE level. Special logging case here since dumping notebook
        # contents can clog the log
        if is_log_level_trace_from_executor():
            self.logger.debug(f'executed notebook contents from {self.merged_notebook_path}:{nb_notebook_merged}')

        # Remove login cell from notebook
        del nb_notebook_merged['cells'][0]

        # Decrement the "execution_count" by 1 to correct notebook cell numbering
        # "execution_count" can have 'None' as value so just pass on any exception to continue
        for cell in nb_notebook_merged['cells']:
            # noinspection PyBroadException
            try:
                if 'execution_count' in cell:
                    execution_count = int(cell['execution_count'])
                    execution_count -= 1
                    cell['execution_count'] = execution_count
            except BaseException:
                pass

            if 'outputs' in cell:
                for output in cell['outputs']:
                    # noinspection PyBroadException
                    try:
                        if 'execution_count' in output:
                            execution_count = int(output['execution_count'])
                            execution_count -= 1
                            output['execution_count'] = execution_count
                    except BaseException:
                        pass

        # Write the scheduled notebook
        with open(self.merged_notebook_path, 'w') as f_notebook_merged:
            nbformat.write(nb_notebook_merged, f_notebook_merged)

        # Log to executor
        self.logger.debug('successfully edited merged notebook')

        # The executed notebook will be returned if spy.jobs.execute was called by a user
        return nb_notebook_merged

    def export_html(self):
        # Open the modified notebook that has been scheduled for execution
        with open(self.merged_notebook_path) as f_notebook_merged:
            nb_notebook_merged = nbformat.read(f_notebook_merged, nbformat.NO_CONVERT)

        # Configure the HTML exporter and export
        html_exporter = nbconvert.HTMLExporter()
        job_result_html, _ = html_exporter.from_notebook_node(nb_notebook_merged)

        # Create parent folder if not existing and write out the exported html to file
        Path(self.job_result_path).parent.mkdir(parents=True, exist_ok=True)
        with open(self.job_result_path, 'w') as f_job_result_file:
            f_job_result_file.write(job_result_html)

        # Log to executor
        self.logger.debug('successfully exported merged notebook')


def execute():
    """
    Execute a notebook. (Internal Seeq function: Not intended for end-users)
    """
    return ExecutionInstance().execute()


def _compose_filename(index, label, scheduled_file_filename, scheduled_file_folder, extension='.html', result=True):
    folder_path: Union[str, Path] = Path(scheduled_file_folder, RESULTS_FOLDER) if result else \
        scheduled_file_folder
    hidden_file_prefix = '.' if not result else ''
    filename_no_ext: str = hidden_file_prefix + Path(scheduled_file_filename).stem

    # Build up the job result html file name
    folder = str(folder_path)
    executor = ".executor"
    index = '.' + index if len(index) > 0 else ''
    label = '.' + label if len(label) > 0 else ''

    result_filename = f'{folder}/{filename_no_ext}{executor}{index}{label}{extension}'
    return result_filename


def get_log_level_from_executor():
    return str(os.environ.get('LOG_LEVEL', 'INFO')).upper()


def is_log_level_trace_from_executor():
    return get_log_level_from_executor() == "TRACE"



from typing import Optional

import pandas as pd
from seeq import spy
from seeq.spy import _common
from seeq.spy._errors import *


class ItemMap:
    """
    Represents a map of item identifiers from a source (usually a saved set of workbooks) to the destination on the
    Seeq Server. This class is used extensively by spy.workbooks.push() operations.
    """

    _lookup_df: pd.DataFrame

    def __init__(self, item_map=None, lookup_df=None):
        self._item_map = item_map if item_map is not None else dict()

        if lookup_df is None:
            self._lookup_df = pd.DataFrame()
        else:
            self._lookup_df = lookup_df

    def __contains__(self, key):
        key = _common.ensure_upper_case_id('ID', key)
        return self._item_map.__contains__(key)

    def __getitem__(self, key):
        key = _common.ensure_upper_case_id('ID', key)
        return self._item_map.__getitem__(key)

    def __setitem__(self, key, val):
        key = _common.ensure_upper_case_id('ID', key)
        val = _common.ensure_upper_case_id('ID', val)
        self._item_map.__setitem__(key, val)

    def __delitem__(self, key):
        key = _common.ensure_upper_case_id('ID', key)
        self._item_map.__delitem__(key)

    def keys(self):
        return self._item_map.keys()

    def look_up_id(self, value):
        row = _common.look_up_in_df(value, self._lookup_df)
        return row['ID'] if row is not None else None


class OverrideItemMap(ItemMap):
    """
    Takes an existing ItemMap and overrides various keys. This is used extensively in the templating system to
    temporarily (for the extent of a particular frame in a callstack) override the inner ItemMap using
    template_parameters or an override_map.
    """

    _override: dict
    _parameters: Optional[dict]

    def __init__(self, item_map: ItemMap, *, template_parameters: dict = None, override_map: dict = None):
        super().__init__(item_map)

        self._override = dict() if override_map is None else override_map
        self._parameters = None
        if template_parameters:
            self._parameters = template_parameters
            self._override_from_template_parameters(template_parameters)

    def __contains__(self, key):
        key = _common.ensure_upper_case_id('ID', key)
        return self._override.__contains__(key) or self._item_map.__contains__(key)

    def __getitem__(self, key):
        key = _common.ensure_upper_case_id('ID', key)
        if self._override.__contains__(key):
            return self._override.__getitem__(key)
        else:
            return super().__getitem__(key)

    def __delitem__(self, key):
        key = _common.ensure_upper_case_id('ID', key)
        if self._override.__contains__(key):
            self._override.__delitem__(key)
        else:
            self._item_map.__delitem__(key)

    def look_up_id(self, value):
        return self._item_map.look_up_id(value)

    def keys(self):
        keys = set(self._item_map.keys())
        keys.update(self._override.keys())
        return keys

    def override(self, key, value):
        self._override[key] = value

    @property
    def parameters(self):
        if self._parameters is not None:
            return self._parameters

        if isinstance(self._item_map, OverrideItemMap):
            return self._item_map.parameters

        return None

    def _override_from_template_parameters(self, template_parameters: dict):

        for key, value in template_parameters.items():
            if value is None:
                continue

            _id, _type, _fqn = spy.workbooks.ItemTemplate.code_key_tuple(key)

            if _id is None:
                # This is the case of a Mustachioed annotation with {{variable}} tokens in it
                continue

            if isinstance(value, pd.DataFrame):
                if len(value) > 1:
                    raise SPyValueError(f'Multiple rows in template_parameters dict for "{key}":\n{value}')
                value = value.iloc[0].to_dict()
            elif isinstance(value, str):
                value = {'Name': value}
            elif isinstance(value, spy.workbooks.ItemTemplate):
                continue

            if _common.present(value, 'ID') and not _common.get(value, 'Reference', default=False):
                self._override[_id] = _common.get(value, 'ID')
            else:
                self._override[_id] = self.look_up_id(value)



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 58.1.3-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from pprint import pformat
from six import iteritems
import re


class ScheduledNotebookInputV1(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'description': 'str',
        'enabled': 'bool',
        'file_path': 'str',
        'label': 'str',
        'schedules': 'list[ScheduleInputV1]',
        'timezone': 'str',
        'user_id': 'str'
    }

    attribute_map = {
        'description': 'description',
        'enabled': 'enabled',
        'file_path': 'filePath',
        'label': 'label',
        'schedules': 'schedules',
        'timezone': 'timezone',
        'user_id': 'userId'
    }

    def __init__(self, description=None, enabled=True, file_path=None, label=None, schedules=None, timezone=None, user_id=None):
        """
        ScheduledNotebookInputV1 - a model defined in Swagger
        """

        self._description = None
        self._enabled = None
        self._file_path = None
        self._label = None
        self._schedules = None
        self._timezone = None
        self._user_id = None

        if description is not None:
          self.description = description
        if enabled is not None:
          self.enabled = enabled
        if file_path is not None:
          self.file_path = file_path
        if label is not None:
          self.label = label
        if schedules is not None:
          self.schedules = schedules
        if timezone is not None:
          self.timezone = timezone
        if user_id is not None:
          self.user_id = user_id

    @property
    def description(self):
        """
        Gets the description of this ScheduledNotebookInputV1.
        Clarifying information or other plain language description of this item. An input of just whitespace is equivalent to a null input

        :return: The description of this ScheduledNotebookInputV1.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """
        Sets the description of this ScheduledNotebookInputV1.
        Clarifying information or other plain language description of this item. An input of just whitespace is equivalent to a null input

        :param description: The description of this ScheduledNotebookInputV1.
        :type: str
        """

        self._description = description

    @property
    def enabled(self):
        """
        Gets the enabled of this ScheduledNotebookInputV1.
        Whether the schedule should be enabled to run jobs

        :return: The enabled of this ScheduledNotebookInputV1.
        :rtype: bool
        """
        return self._enabled

    @enabled.setter
    def enabled(self, enabled):
        """
        Sets the enabled of this ScheduledNotebookInputV1.
        Whether the schedule should be enabled to run jobs

        :param enabled: The enabled of this ScheduledNotebookInputV1.
        :type: bool
        """

        self._enabled = enabled

    @property
    def file_path(self):
        """
        Gets the file_path of this ScheduledNotebookInputV1.
        The path to the ipynb file to be run (case-sensitive). May not be altered after initial creation

        :return: The file_path of this ScheduledNotebookInputV1.
        :rtype: str
        """
        return self._file_path

    @file_path.setter
    def file_path(self, file_path):
        """
        Sets the file_path of this ScheduledNotebookInputV1.
        The path to the ipynb file to be run (case-sensitive). May not be altered after initial creation

        :param file_path: The file_path of this ScheduledNotebookInputV1.
        :type: str
        """
        if file_path is None:
            raise ValueError("Invalid value for `file_path`, must not be `None`")

        self._file_path = file_path

    @property
    def label(self):
        """
        Gets the label of this ScheduledNotebookInputV1.
        An optional label that differentiates this schedule (case-sensitive), allowing the samefilePath to be run by multiple users. May not be altered after initial creation

        :return: The label of this ScheduledNotebookInputV1.
        :rtype: str
        """
        return self._label

    @label.setter
    def label(self, label):
        """
        Sets the label of this ScheduledNotebookInputV1.
        An optional label that differentiates this schedule (case-sensitive), allowing the samefilePath to be run by multiple users. May not be altered after initial creation

        :param label: The label of this ScheduledNotebookInputV1.
        :type: str
        """

        self._label = label

    @property
    def schedules(self):
        """
        Gets the schedules of this ScheduledNotebookInputV1.
        The Notebook's run schedule.  Any entries for which a key is not provided will be supplied with a key from a 0-based list of integers

        :return: The schedules of this ScheduledNotebookInputV1.
        :rtype: list[ScheduleInputV1]
        """
        return self._schedules

    @schedules.setter
    def schedules(self, schedules):
        """
        Sets the schedules of this ScheduledNotebookInputV1.
        The Notebook's run schedule.  Any entries for which a key is not provided will be supplied with a key from a 0-based list of integers

        :param schedules: The schedules of this ScheduledNotebookInputV1.
        :type: list[ScheduleInputV1]
        """

        self._schedules = schedules

    @property
    def timezone(self):
        """
        Gets the timezone of this ScheduledNotebookInputV1.
        The timezone in which the scheduled times will be run, defaults to UTC

        :return: The timezone of this ScheduledNotebookInputV1.
        :rtype: str
        """
        return self._timezone

    @timezone.setter
    def timezone(self, timezone):
        """
        Sets the timezone of this ScheduledNotebookInputV1.
        The timezone in which the scheduled times will be run, defaults to UTC

        :param timezone: The timezone of this ScheduledNotebookInputV1.
        :type: str
        """

        self._timezone = timezone

    @property
    def user_id(self):
        """
        Gets the user_id of this ScheduledNotebookInputV1.
        The user ID to run the Notebook as, defaults to the current user if not specified. Only an admin user is allowed to specify a user ID

        :return: The user_id of this ScheduledNotebookInputV1.
        :rtype: str
        """
        return self._user_id

    @user_id.setter
    def user_id(self, user_id):
        """
        Sets the user_id of this ScheduledNotebookInputV1.
        The user ID to run the Notebook as, defaults to the current user if not specified. Only an admin user is allowed to specify a user ID

        :param user_id: The user_id of this ScheduledNotebookInputV1.
        :type: str
        """

        self._user_id = user_id

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, ScheduledNotebookInputV1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



import json
import re

import chevron
from seeq.spy import _common


class MustachioedAnnotation:
    """
    Wraps a piece of HTML and exposes any Mustache variables in a way that is useful to the templating system. Also
    facilitates rendering of a template with Mustache variables replaced.
    """

    INVALID_MUSTACHE_CHARS = r'\.'
    INVALID_MUSTACHE_CHARS_REGEX = re.compile('[' + INVALID_MUSTACHE_CHARS + ']')

    _html: str
    _code_dict: dict

    def __init__(self, html, resolve_content_key_func=None):
        self._html = _common.fix_up_ckeditor_curly_brace_weirdness(html)

        if resolve_content_key_func is not None:
            self._html = MustachioedAnnotation._parameterize_seeq_content_ids(self._html, resolve_content_key_func)

        self._validate()

    @property
    def html(self):
        return self._html

    @property
    def code(self):
        return json.dumps(self._code_dict, indent=4).replace(': null', ': None')

    @property
    def code_dict(self):
        return self._code_dict

    @staticmethod
    def _parameterize_seeq_content_ids(html, resolve_content_key_func):
        return re.sub(r'data-seeq-content="([^"]+)"',
                      lambda m: f'data-seeq-content="{{{{{resolve_content_key_func(m.group(1))}}}}}"',
                      html)

    # noinspection PyBroadException
    @staticmethod
    def invalid_mustache_characters():
        invalid_chars = str()
        for i in range(0, 65536):

            try:
                output = chevron.render('{{z' + chr(i) + 'z}}', {'z' + chr(i) + 'z': 'Worked!'})
                if output != 'Worked!':
                    invalid_chars += chr(i)
            except BaseException:
                invalid_chars += chr(i)

        return invalid_chars

    @staticmethod
    def _sanitize_mustache_token(token):
        return MustachioedAnnotation.INVALID_MUSTACHE_CHARS_REGEX.sub('-', token)

    def _validate(self):
        matches = re.finditer(r'{{([#/^]?)([^}]+)}}', self.html)
        replacements = list()
        root_dict = dict()
        stack = list([root_dict])
        current_dict = root_dict
        for match in matches:
            qualifier = match.group(1)
            name = match.group(2)

            if name is None:
                continue

            sanitized_name = MustachioedAnnotation._sanitize_mustache_token(name)

            if qualifier is not None:
                replacements.append((match.group(0), '{{' + qualifier + sanitized_name + '}}'))
            else:
                replacements.append((match.group(0), '{{' + sanitized_name + '}}'))

            new_list = [dict()]
            if qualifier in ['#', '^']:
                stack.append(current_dict)
                if sanitized_name in current_dict:
                    current_dict = current_dict[sanitized_name][0]
                else:
                    current_dict[sanitized_name] = new_list
                    current_dict = new_list[0]
            elif qualifier == '/':
                current_dict = stack.pop()
            else:
                current_dict[sanitized_name] = None

        for _old, _new in replacements:
            self._html = self._html.replace(_old, _new)

        self._code_dict = root_dict

    def render(self, parameters: dict) -> str:
        return chevron.render(self.html, parameters)



from seeq.spy._common import \
    is_guid, \
    time_abbreviation_to_ms_multiplier, \
    parse_str_time_to_ms
from seeq.spy._login import \
    get_user_timezone, \
    is_valid_unit
from seeq.spy._url import \
    get_workbook_id_from_url, \
    get_worksheet_id_from_url, \
    get_workstep_id_from_url, \
    get_data_lab_project_id_from_url
from seeq.spy._datalab import get_notebook_url, get_data_lab_project_url, get_data_lab_project_id
from seeq.spy.workbooks._worksheet import get_analysis_worksheet_from_url

__all__ = ['get_workbook_id_from_url',
           'get_worksheet_id_from_url',
           'get_workstep_id_from_url',
           'get_data_lab_project_id_from_url',
           'get_analysis_worksheet_from_url',
           'is_guid',
           'is_valid_unit',
           'time_abbreviation_to_ms_multiplier',
           'parse_str_time_to_ms',
           'get_data_lab_project_url',
           'get_data_lab_project_id',
           'get_user_timezone',
           'get_notebook_url']



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 58.1.3-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from pprint import pformat
from six import iteritems
import re


class DatasourceInputV1(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'additional_properties': 'list[ScalarPropertyV1]',
        'cache_enabled': 'bool',
        'condition_location': 'str',
        'data_version_check': 'str',
        'datasource_class': 'str',
        'datasource_id': 'str',
        'description': 'str',
        'indexing_schedule_supported': 'bool',
        'name': 'str',
        'signal_location': 'str',
        'stored_in_seeq': 'bool'
    }

    attribute_map = {
        'additional_properties': 'additionalProperties',
        'cache_enabled': 'cacheEnabled',
        'condition_location': 'conditionLocation',
        'data_version_check': 'dataVersionCheck',
        'datasource_class': 'datasourceClass',
        'datasource_id': 'datasourceId',
        'description': 'description',
        'indexing_schedule_supported': 'indexingScheduleSupported',
        'name': 'name',
        'signal_location': 'signalLocation',
        'stored_in_seeq': 'storedInSeeq'
    }

    def __init__(self, additional_properties=None, cache_enabled=False, condition_location=None, data_version_check=None, datasource_class=None, datasource_id=None, description=None, indexing_schedule_supported=False, name=None, signal_location=None, stored_in_seeq=False):
        """
        DatasourceInputV1 - a model defined in Swagger
        """

        self._additional_properties = None
        self._cache_enabled = None
        self._condition_location = None
        self._data_version_check = None
        self._datasource_class = None
        self._datasource_id = None
        self._description = None
        self._indexing_schedule_supported = None
        self._name = None
        self._signal_location = None
        self._stored_in_seeq = None

        if additional_properties is not None:
          self.additional_properties = additional_properties
        if cache_enabled is not None:
          self.cache_enabled = cache_enabled
        if condition_location is not None:
          self.condition_location = condition_location
        if data_version_check is not None:
          self.data_version_check = data_version_check
        if datasource_class is not None:
          self.datasource_class = datasource_class
        if datasource_id is not None:
          self.datasource_id = datasource_id
        if description is not None:
          self.description = description
        if indexing_schedule_supported is not None:
          self.indexing_schedule_supported = indexing_schedule_supported
        if name is not None:
          self.name = name
        if signal_location is not None:
          self.signal_location = signal_location
        if stored_in_seeq is not None:
          self.stored_in_seeq = stored_in_seeq

    @property
    def additional_properties(self):
        """
        Gets the additional_properties of this DatasourceInputV1.
        Additional properties to set on the datasource

        :return: The additional_properties of this DatasourceInputV1.
        :rtype: list[ScalarPropertyV1]
        """
        return self._additional_properties

    @additional_properties.setter
    def additional_properties(self, additional_properties):
        """
        Sets the additional_properties of this DatasourceInputV1.
        Additional properties to set on the datasource

        :param additional_properties: The additional_properties of this DatasourceInputV1.
        :type: list[ScalarPropertyV1]
        """

        self._additional_properties = additional_properties

    @property
    def cache_enabled(self):
        """
        Gets the cache_enabled of this DatasourceInputV1.
        True if this datasource's signal data should be cached in Seeq; false otherwise

        :return: The cache_enabled of this DatasourceInputV1.
        :rtype: bool
        """
        return self._cache_enabled

    @cache_enabled.setter
    def cache_enabled(self, cache_enabled):
        """
        Sets the cache_enabled of this DatasourceInputV1.
        True if this datasource's signal data should be cached in Seeq; false otherwise

        :param cache_enabled: The cache_enabled of this DatasourceInputV1.
        :type: bool
        """

        self._cache_enabled = cache_enabled

    @property
    def condition_location(self):
        """
        Gets the condition_location of this DatasourceInputV1.
        If StoredInSeeq is set to true, then this is where conditions under this datasource will be stored. Valid values: POSTGRES.

        :return: The condition_location of this DatasourceInputV1.
        :rtype: str
        """
        return self._condition_location

    @condition_location.setter
    def condition_location(self, condition_location):
        """
        Sets the condition_location of this DatasourceInputV1.
        If StoredInSeeq is set to true, then this is where conditions under this datasource will be stored. Valid values: POSTGRES.

        :param condition_location: The condition_location of this DatasourceInputV1.
        :type: str
        """
        allowed_values = ["FILE_SIGNAL_CACHE", "FILE_SIGNAL_STORAGE", "POSTGRES"]
        if condition_location not in allowed_values:
            raise ValueError(
                "Invalid value for `condition_location` ({0}), must be one of {1}"
                .format(condition_location, allowed_values)
            )

        self._condition_location = condition_location

    @property
    def data_version_check(self):
        """
        Gets the data_version_check of this DatasourceInputV1.
        The datasource's version check, which represents a version of the metadata synced for this datasource. If its data version check is unchanged on subsequent metadata syncs, no update is necessary. A common data version check is a datetime, checksum or last change date.

        :return: The data_version_check of this DatasourceInputV1.
        :rtype: str
        """
        return self._data_version_check

    @data_version_check.setter
    def data_version_check(self, data_version_check):
        """
        Sets the data_version_check of this DatasourceInputV1.
        The datasource's version check, which represents a version of the metadata synced for this datasource. If its data version check is unchanged on subsequent metadata syncs, no update is necessary. A common data version check is a datetime, checksum or last change date.

        :param data_version_check: The data_version_check of this DatasourceInputV1.
        :type: str
        """

        self._data_version_check = data_version_check

    @property
    def datasource_class(self):
        """
        Gets the datasource_class of this DatasourceInputV1.
        The class of the datasource

        :return: The datasource_class of this DatasourceInputV1.
        :rtype: str
        """
        return self._datasource_class

    @datasource_class.setter
    def datasource_class(self, datasource_class):
        """
        Sets the datasource_class of this DatasourceInputV1.
        The class of the datasource

        :param datasource_class: The datasource_class of this DatasourceInputV1.
        :type: str
        """
        if datasource_class is None:
            raise ValueError("Invalid value for `datasource_class`, must not be `None`")

        self._datasource_class = datasource_class

    @property
    def datasource_id(self):
        """
        Gets the datasource_id of this DatasourceInputV1.
        The ID of the datasource

        :return: The datasource_id of this DatasourceInputV1.
        :rtype: str
        """
        return self._datasource_id

    @datasource_id.setter
    def datasource_id(self, datasource_id):
        """
        Sets the datasource_id of this DatasourceInputV1.
        The ID of the datasource

        :param datasource_id: The datasource_id of this DatasourceInputV1.
        :type: str
        """
        if datasource_id is None:
            raise ValueError("Invalid value for `datasource_id`, must not be `None`")

        self._datasource_id = datasource_id

    @property
    def description(self):
        """
        Gets the description of this DatasourceInputV1.
        Clarifying information or other plain language description of this item. An input of just whitespace is equivalent to a null input.

        :return: The description of this DatasourceInputV1.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """
        Sets the description of this DatasourceInputV1.
        Clarifying information or other plain language description of this item. An input of just whitespace is equivalent to a null input.

        :param description: The description of this DatasourceInputV1.
        :type: str
        """

        self._description = description

    @property
    def indexing_schedule_supported(self):
        """
        Gets the indexing_schedule_supported of this DatasourceInputV1.
        True if for this datasource indexing can be scheduled or requested from Datasource Management UI.

        :return: The indexing_schedule_supported of this DatasourceInputV1.
        :rtype: bool
        """
        return self._indexing_schedule_supported

    @indexing_schedule_supported.setter
    def indexing_schedule_supported(self, indexing_schedule_supported):
        """
        Sets the indexing_schedule_supported of this DatasourceInputV1.
        True if for this datasource indexing can be scheduled or requested from Datasource Management UI.

        :param indexing_schedule_supported: The indexing_schedule_supported of this DatasourceInputV1.
        :type: bool
        """

        self._indexing_schedule_supported = indexing_schedule_supported

    @property
    def name(self):
        """
        Gets the name of this DatasourceInputV1.
        Human readable name. Required during creation. An input of just whitespaces is equivalent to a null input.

        :return: The name of this DatasourceInputV1.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this DatasourceInputV1.
        Human readable name. Required during creation. An input of just whitespaces is equivalent to a null input.

        :param name: The name of this DatasourceInputV1.
        :type: str
        """
        if name is None:
            raise ValueError("Invalid value for `name`, must not be `None`")

        self._name = name

    @property
    def signal_location(self):
        """
        Gets the signal_location of this DatasourceInputV1.
        If StoredInSeeq is set to true, then this is where signals under this datasource will be stored. Valid values: FILE_SIGNAL_STORAGE.

        :return: The signal_location of this DatasourceInputV1.
        :rtype: str
        """
        return self._signal_location

    @signal_location.setter
    def signal_location(self, signal_location):
        """
        Sets the signal_location of this DatasourceInputV1.
        If StoredInSeeq is set to true, then this is where signals under this datasource will be stored. Valid values: FILE_SIGNAL_STORAGE.

        :param signal_location: The signal_location of this DatasourceInputV1.
        :type: str
        """
        allowed_values = ["FILE_SIGNAL_CACHE", "FILE_SIGNAL_STORAGE", "POSTGRES"]
        if signal_location not in allowed_values:
            raise ValueError(
                "Invalid value for `signal_location` ({0}), must be one of {1}"
                .format(signal_location, allowed_values)
            )

        self._signal_location = signal_location

    @property
    def stored_in_seeq(self):
        """
        Gets the stored_in_seeq of this DatasourceInputV1.
        True if this datasource's data is stored in Seeq; false otherwise (e.g. a remote datasource like PI).

        :return: The stored_in_seeq of this DatasourceInputV1.
        :rtype: bool
        """
        return self._stored_in_seeq

    @stored_in_seeq.setter
    def stored_in_seeq(self, stored_in_seeq):
        """
        Sets the stored_in_seeq of this DatasourceInputV1.
        True if this datasource's data is stored in Seeq; false otherwise (e.g. a remote datasource like PI).

        :param stored_in_seeq: The stored_in_seeq of this DatasourceInputV1.
        :type: bool
        """

        self._stored_in_seeq = stored_in_seeq

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, DatasourceInputV1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



from __future__ import annotations

import queue
import re
from threading import Event, Thread
from typing import Callable

import pandas as pd

from seeq.spy import _common

SCOPED_DATA_ID_REGEX = re.compile(r'^\[(.*?)] \{(.*?)} (.*)$')


class PushResults(dict):
    _by_asset_index: dict
    _by_workbook_and_path_index: dict
    _by_data_id_index: dict

    _post_thread: Thread
    _post_thread_shut_down: bool
    _post_thread_has_shut_down: Event
    _post_queue: queue.Queue
    _response_queue: queue.Queue

    flush_section_count: int

    POST_QUEUE_SIZE = 8

    def __init__(self, df: pd.DataFrame):
        self._by_asset_index = dict()
        self._by_workbook_and_path_index = dict()
        self._by_data_id_index = dict()
        self._post_queue = queue.Queue(maxsize=PushResults.POST_QUEUE_SIZE)
        self._response_queue = queue.Queue()
        self._flush_section_count = 0

        super().__init__({index: PushItem(index, row.to_dict(), self._set_item_callback)
                          for index, row in df.iterrows()})

    def _post_thread_main_loop(self):
        while True:
            if self._post_thread_shut_down and self._post_queue.empty():
                break

            try:
                func, args = self._post_queue.get(block=True, timeout=0.1)
                func(*args)
            except queue.Empty:
                pass
            except BaseException as e:
                self._response_queue.put((PushResults._raise_exception, (e,)))

        self._post_thread_has_shut_down.set()

    def add_post(self, func, args):
        self._post_queue.put((func, args))

    def add_response(self, func, args):
        self._response_queue.put((func, args))

    @staticmethod
    def _raise_exception(e):
        raise e

    def drain_responses(self):
        while True:
            try:
                func, args = self._response_queue.get_nowait()
                func(*args)
            except queue.Empty:
                break

    def start_post_thread(self):
        self._post_thread_shut_down = False
        self._post_thread_has_shut_down = Event()
        self._post_thread = Thread(target=self._post_thread_main_loop)
        self._post_thread.start()

    def shut_down_post_thread(self):
        self._post_thread_shut_down = True
        self._post_thread_has_shut_down.wait()

    def __setitem__(self, key, value):
        if not isinstance(value, dict):
            raise ValueError('value must be dict')

        if not isinstance(value, PushItem):
            value = PushItem(key, value, self._set_item_callback)
        else:
            self._set_item_callback(key, value, None)

        value.callback = self._set_item_callback

        super().__setitem__(key, value)

    def __getitem__(self, item):
        return super().__getitem__(item)

    def __contains__(self, item):
        return super().__contains__(item)

    def _set_item_callback(self, index, item, key):
        if key not in [None, 'Type', 'Name', 'Path', 'Data ID']:
            return

        self._update_indices(index, item)

    def _update_indices(self, index, item):
        if _common.get(item, 'Type') == 'Asset':
            self._by_asset_index[(_common.get(item, 'Name'), _common.get(item, 'Path'))] = index

        if _common.present(item, 'Data ID'):
            self._by_data_id_index[item['Data ID']] = index
            matcher = SCOPED_DATA_ID_REGEX.match(item['Data ID'])
            if matcher is not None:
                workbook_id = matcher.group(1)
                path = matcher.group(3)
                self._by_workbook_and_path_index[(workbook_id, path)] = index

    def get_by_asset(self, name, path):
        return self._by_asset_index.get((name, path))

    def get_by_workbook_and_path(self, workbook_id, path):
        return self._by_workbook_and_path_index.get((workbook_id, path))

    def get_by_data_id(self, data_id):
        return self._by_data_id_index.get(data_id)

    class AtIndexer:
        _parent: PushResults

        def __init__(self, push_results):
            self._parent = push_results

        def __getitem__(self, item):
            return self._parent[item[0]][item[1]]

        def __setitem__(self, key, value):
            self._parent[key[0]][key[1]] = value

        def __contains__(self, item):
            return item[0] in self._parent and item[1] in self._parent[item[0]]

    class LocIndexer:
        _parent: PushResults

        def __init__(self, push_results):
            self._parent = push_results

        def __getitem__(self, item):
            return self._parent[item]

        def __setitem__(self, key, value):
            self._parent[key] = value

        def __contains__(self, item):
            return item in self._parent

    @property
    def at(self):
        return PushResults.AtIndexer(self)

    @property
    def loc(self):
        return PushResults.LocIndexer(self)


class PushItem(dict):
    _index: object
    _callback: Callable

    def __init__(self, index, value, callback):
        super().__init__(value)
        self._index = index
        self._callback = callback
        self._callback(index, value, None)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self._callback(self._index, self, key)

    def __getitem__(self, item):
        return super().__getitem__(item)

    def __contains__(self, item):
        return super().__contains__(item)

    def update(self, other, **kwargs):
        super().update(other, **kwargs)
        self._callback(self._index, self, None)



from __future__ import annotations

import copy
import json
import re
from typing import Optional, List, Callable, Union

from seeq.spy import _common
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.workbooks._annotation import Annotation, Journal, Report
from seeq.spy.workbooks._content import DateRange, AssetSelection, Content
from seeq.spy.workbooks._item import Item, ItemList, Reference
from seeq.spy.workbooks._item_map import ItemMap, OverrideItemMap
from seeq.spy.workbooks._mustache import MustachioedAnnotation
from seeq.spy.workbooks._workbook import Workbook, Analysis, Topic
from seeq.spy.workbooks._worksheet import Worksheet, AnalysisWorksheet, ExistingWorkstepWalk, TopicDocument
from seeq.spy.workbooks._workstep import AnalysisWorkstep


#
# This templating system allows a user to define an Analysis or Topic in Workbench or Organizer, save it to disk and
# then use it as a template whereby the items within a worksheet/workstep/date range are replaced programmatically or
# the context in a Topic Document is replaced with references to a worksheet/workstep.
#
# This file contains a series of XxxxxTemplate classes that wrap an "inner" Xxxxx instance object that serves as the
# template. The wrapper class provides the same interface as the inner object but does some work to substitute/override
# items, worksheets/worksteps, and Mustache variables before pushing to Seeq Server.
#
# If the user sets any values on the definition, it goes into the _definition_override. The inner object is never
# touched.
#
# Every template is necessarily created with a label, which is used to differentiate between multiple instances of a
# template. This allows the normal spy.workbooks.push() operation to properly keep track of identifiers and keep
# idempotency.
#
# Each XxxxxTemplate class exposes a "code" attribute which provides boilerplate code that the user copy/pastes in
# order to override the template "parameters".
#
# Template instances are typically created using the spy.workbooks.load(as_template_with_label) argument.
#
# See "Workbook Templates.ipynb" for a tutorial on how to use.
#

class ItemTemplate:
    _definition_override: dict
    _label: str
    _parameters: Optional[dict]
    _template: Item
    _is_copy: bool

    PARAMETER_KEY_REGEX = re.compile(
        r'(?P<id>[0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12})?'
        r'\s*(\[(?P<type>[^]]+)])?(\s*(?P<fqn>.*))?'
    )

    def __init__(self, label: str, template: Item, package: ItemList, is_copy: bool):
        if isinstance(template, ItemTemplate):
            raise SPyTypeError('template parameter cannot be a XxxxxTemplate object, it must be a "regular" object')

        self._definition_override = dict()
        self._label = label
        self._parameters = None
        self._template = template
        self._package = package if package is not None else ItemList()
        self._is_copy = is_copy

        # This one of the key concepts of the template: The ID of a template is the combination of the wrapped item
        # ID and the label for the template that is supplied by the user.
        self._definition_override['ID'] = f'{template.id} {self._label}'

    #
    # The following dictionary-like indexing functions facilitate the "override" mechanism that the template provides
    #

    def __contains__(self, key):
        return _common.present(self._definition_override, key) or key in self._template

    def __getitem__(self, key):
        return self._definition_override[key] if key in self._definition_override else self._template[key]

    def __setitem__(self, key, val):
        self._definition_override[key] = _common.ensure_upper_case_id(key, val)

    def __delitem__(self, key):
        del self._definition_override[key]

    def __repr__(self):
        return '%s "%s" (%s)' % (self.type, self.fqn, self.id)

    @property
    def id(self):
        return self['ID']

    @property
    def name(self):
        return self['Name']

    @name.setter
    def name(self, val):
        self._definition_override['Name'] = val

    @property
    def type(self):
        return self['Type']

    def refresh_from(self, new_item, item_map: ItemMap, status: Status):
        # We don't allow refreshing of Template items because it's just too confusing... users would not expect to
        # have a Template object trying to directly represent something on the server
        pass

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, val):
        self._label = val

    @property
    def template(self):
        return self._template

    @property
    def is_copy(self):
        return self._is_copy

    @property
    def package(self):
        return self._package

    @property
    def parameters(self) -> dict:
        return self._parameters

    @parameters.setter
    def parameters(self, val: dict):
        self._set_parameters(val)

    def _validate_parameters_dict(self, code_dict, parameters_dict):
        final = code_dict.copy()
        for k, v in parameters_dict.items():
            if k not in code_dict:
                matching_keys = list()
                regex_match = ItemTemplate.PARAMETER_KEY_REGEX.match(k)
                if not regex_match:
                    raise SPyValueError(f'parameters key not recognized: "{k}"\nKey format must be "ID [Type] FQN"')

                _id = regex_match.group('id')
                _type = regex_match.group('type')
                _fqn = regex_match.group('fqn')

                for code_key_string in code_dict.keys():
                    code_key = ItemTemplate.code_key_tuple(code_key_string)
                    if _id is not None and _id == code_key[0]:
                        matching_keys.append(code_key_string)
                    elif (_fqn is not None and _fqn == code_key[2]) and (_type is None or _type == code_key[1]):
                        matching_keys.append(code_key_string)

                if len(matching_keys) == 0:
                    raise SPyValueError(
                        f'parameters key "{k}" could not be mapped to anything in the template code:\n' +
                        self.code)

                if len(matching_keys) >= 2:
                    raise SPyValueError(f'parameters key "{k}" matches multiple keys in the template code:\n' +
                                        '\n'.join([k for k in matching_keys]))

                k = matching_keys[0]

            if isinstance(code_dict[k], list):
                if not isinstance(code_dict[k], type(v)):
                    raise SPyValueError(f'parameters key "{k}" is a {type(v)}, should be a {type(code_dict[k])}')

                if len(code_dict[k]) == 0:
                    raise SPyValueError(f'parameters key "{k}" should be omitted as there are no nested moustache '
                                        f'elements present in the Topic Document for it')

                final[k] = self._validate_parameters_list(code_dict[k][0], v)
            elif isinstance(code_dict[k], dict):
                final[k] = self._validate_parameters_dict(code_dict[k], v)
            else:
                final[k] = v

        return final

    def _validate_parameters_list(self, code_dict_for_list, parameters_list):
        final = list()

        for list_item in parameters_list:
            if isinstance(list_item, dict):
                final.append(self._validate_parameters_dict(code_dict_for_list, list_item))
            else:
                final.append(list_item)

        return final

    def _set_parameters(self, parameters):
        if parameters is None:
            self._parameters = None
            return

        self._parameters = self._validate_parameters_dict(self.code_dict(), parameters)

    @property
    def provenance(self):
        return Item.TEMPLATE

    @property
    def fqn(self):
        parts = list()
        if _common.present(self, 'Path'):
            parts.append(self['Path'])
        if self.type != 'Asset' and _common.present(self, 'Asset'):
            parts.append(self['Asset'])
        if _common.present(self, 'Name'):
            parts.append(self['Name'])

        return ' >> '.join(parts)

    @property
    def definition_dict(self):
        d = copy.deepcopy(self._template.definition_dict)
        d.update(self._definition_override)
        return d

    @staticmethod
    def code_key_string(_id, _type, _fqn):
        return f'{_id} [{_type}] {_fqn}'

    @staticmethod
    def code_key_tuple(s):
        match = ItemTemplate.PARAMETER_KEY_REGEX.match(s)
        return match.group('id'), match.group('type'), match.group('fqn')

    @staticmethod
    def _get_template_code_string(_code_dict):
        return json.dumps(_code_dict, indent=4).replace(': null', ': None')

    def _code_variable(self):
        pass

    def code_dict(self):
        return dict()

    @property
    def code(self):
        return f'{self._code_variable()}.parameters = {ItemTemplate._get_template_code_string(self.code_dict())}'

    def is_overridden(self, key) -> bool:
        return key in self._definition_override

    @staticmethod
    def _code_dict_for_analysis(workbook: Analysis, worksheets: List[Union[AnalysisWorksheet, AnalysisWorkstep]]):
        code_dict = dict()
        for worksheet in worksheets:
            item_ids = {r.id for r in worksheet.referenced_items}
            for item_id in item_ids:
                if item_id not in workbook.item_inventory:
                    # This can happen if the workbook was not pulled with spy.workbooks.pull()
                    obj = Item.pull(item_id)
                else:
                    obj = workbook.item_inventory[item_id]
                clean_fqn = obj.fqn.replace("\"", "")
                clean_type = obj.type.replace("Stored", "").replace("Calculated", "")
                code_dict[ItemTemplate.code_key_string(item_id, clean_type, clean_fqn)] = None

            if not isinstance(worksheet, AnalysisWorkstep):
                mustache = MustachioedAnnotation(
                    worksheet.document.html,
                    lambda content_id: worksheet.document.resolve_content_code_key(content_id))

                code_dict.update(mustache.code_dict)

        return code_dict

    @staticmethod
    def _code_dict_for_topic(documents: List[ReportTemplate]) -> dict:
        code_dict = dict()
        for document in documents:
            for reference in document.referenced_items:
                if reference.id not in document.worksheet.workbook.item_inventory:
                    raise SPyValueError(f'Topic Document {document.worksheet} has '
                                        f'{reference.provenance} with ID "{reference.id}" '
                                        f'but that item is not found in the item inventory of workbook '
                                        f'{document.worksheet.workbook}')

                item = document.worksheet.workbook.item_inventory[reference.id]
                code_dict[ItemTemplate.code_key_string(reference.id, reference.provenance, item.fqn)] = None

            mustache = MustachioedAnnotation(document.html,
                                             lambda content_id: document.resolve_content_code_key(content_id))
            code_dict.update(mustache.code_dict)

        return code_dict

    def _after_push(self, item_map):
        if not self.is_copy:
            # We also map to the non-templated ID in case there's a piece of content that has no template parameters
            item_map[self.template.id] = item_map[self.id]


class WorkbookTemplate(ItemTemplate, Workbook):
    def __init__(self, label: str, workbook: Workbook, *, package: ItemList = None, is_copy: bool):
        if not isinstance(workbook, Workbook):
            raise SPyTypeError('template parameter must be a Workbook (Analysis or Topic)')

        ItemTemplate.__init__(self, label, workbook, package, is_copy)

    @property
    def template(self) -> Workbook:
        # noinspection PyTypeChecker
        return self._template

    @property
    def item_inventory(self):
        return self.template.item_inventory

    @property
    def datasource_maps(self):
        return self.template.datasource_maps

    @property
    def scoped_items(self):
        return self.template.scoped_items

    @property
    def datasource_inventory(self):
        return self.template.datasource_inventory

    def _find_existing_workbook_template(self, workbook_id, label):
        for workbook in self.package:
            if workbook.id == f'{workbook_id} {label}':
                return workbook

        return None

    def push(self, *, owner=None, folder_id=None, item_map: ItemMap = None, label=None, datasource=None,
             access_control=None, override_max_interp=False, include_inventory=True,
             scope_globals_to_workbook=False, specific_worksheet_ids: Optional[List[str]] = None, status=None,
             session: Optional[Session] = None):
        if specific_worksheet_ids is None or len(specific_worksheet_ids) > 0:
            override_item_map = OverrideItemMap(item_map, template_parameters=self.parameters)
        else:
            override_item_map = item_map

        output = super().push(owner=owner, folder_id=folder_id, item_map=override_item_map,
                              label=self.label, datasource=datasource, access_control=access_control,
                              override_max_interp=override_max_interp, include_inventory=include_inventory,
                              scope_globals_to_workbook=scope_globals_to_workbook,
                              specific_worksheet_ids=specific_worksheet_ids, status=status, session=session)

        self._after_push(item_map)

        return output


class AnalysisTemplate(WorkbookTemplate, Analysis):
    def __init__(self, label: str, workbook: Analysis, *, package: ItemList = None, is_copy: bool = False):
        if not isinstance(workbook, Analysis):
            raise SPyTypeError('workbook parameter must be an Analysis')

        WorkbookTemplate.__init__(self, label, workbook, package=package, is_copy=is_copy)
        Analysis.__init__(self)

        # Clear definition for clarity, since it's not used in a template
        self._definition = dict()

        for worksheet in workbook.worksheets:  # type: AnalysisWorksheet
            AnalysisWorksheetTemplate(label, self, worksheet, package=package, is_copy=is_copy)

    @property
    def template(self) -> Analysis:
        # noinspection PyTypeChecker
        return self._template

    def _code_variable(self):
        return 'workbook'

    def code_dict(self):
        return ItemTemplate._code_dict_for_analysis(self.template, self.template.worksheets)

    def copy(self, label):
        existing_workbook = self._find_existing_workbook_template(self.template.id, label)
        if existing_workbook is not None:
            return existing_workbook

        return AnalysisTemplate(label, self.template, package=self.package, is_copy=True)


class TopicTemplate(WorkbookTemplate, Topic):
    def __init__(self, label: str, topic: Topic, *, package: ItemList = None, is_copy: bool = False):
        if not isinstance(topic, Topic):
            raise SPyTypeError('topic parameter must be a Topic')

        WorkbookTemplate.__init__(self, label, topic, package=package, is_copy=is_copy)
        Topic.__init__(self)

        # Clear definition for clarity, since it's not used in a template
        self._definition = dict()

        for worksheet in topic.worksheets:  # type: TopicDocument
            TopicDocumentTemplate(label, self, worksheet, package=package, is_copy=is_copy)

    @property
    def template(self) -> Topic:
        # noinspection PyTypeChecker
        return self._template

    def _code_variable(self):
        return 'topic'

    def code_dict(self):
        return ItemTemplate._code_dict_for_topic([worksheet.document for worksheet in self.worksheets])

    def copy(self, label):
        existing_workbook = self._find_existing_workbook_template(self.template.id, label)
        if existing_workbook is not None:
            return existing_workbook

        return TopicTemplate(label, self.template, package=self.package, is_copy=True)


class WorksheetTemplate(ItemTemplate, Worksheet):
    def __init__(self, label: str, worksheet: Worksheet, package: ItemList, is_copy: bool):
        if not isinstance(worksheet, Worksheet):
            raise SPyTypeError('worksheet parameter must be a Worksheet')

        ItemTemplate.__init__(self, label, worksheet, package, is_copy)

    def _find_existing_worksheet_template(self, worksheet_id, label):
        for worksheet in self.workbook.worksheets:
            if worksheet.id == f'{worksheet_id} {label}':
                return worksheet

        return None


class AnalysisWorksheetTemplate(WorksheetTemplate, AnalysisWorksheet):
    def __init__(self, label: str, workbook, worksheet: AnalysisWorksheet, *, package: ItemList = None,
                 is_copy: bool = False):
        if not isinstance(worksheet, AnalysisWorksheet):
            raise SPyTypeError('worksheet parameter must be a AnalysisWorksheet')

        WorksheetTemplate.__init__(self, label, worksheet, package, is_copy)
        AnalysisWorksheet.__init__(self, workbook, add_defaults=False)

        self._document = JournalTemplate(label, worksheet.document, self, package)

        # Clear definition for clarity, since it's not used in a template
        self._definition = dict()

        self.worksteps = dict()
        for _, workstep in worksheet.worksteps.items():
            AnalysisWorkstepTemplate(label, self, workstep, package=package, is_copy=is_copy)

        self['Current Workstep ID'] = f"{self['Current Workstep ID']} {label}"

    @property
    def template(self) -> AnalysisWorksheet:
        # noinspection PyTypeChecker
        return self._template

    @property
    def parameters(self):
        if self._parameters is None:
            return self.workbook.parameters
        return self._parameters

    @parameters.setter
    def parameters(self, val):
        super()._set_parameters(val)

    def _code_variable(self):
        return 'worksheet'

    def code_dict(self):
        return ItemTemplate._code_dict_for_analysis(self.template.workbook, [self.template])

    @property
    def document(self):
        return self._document

    def push(self, session: Session, pushed_workbook_id, item_map, datasource_output, existing_worksheet_identifiers,
             include_inventory, label=None, status: Status = None):
        output = super().push(session, pushed_workbook_id,
                              OverrideItemMap(item_map, template_parameters=self.parameters), datasource_output,
                              existing_worksheet_identifiers, include_inventory, self.label, status=status)

        self._after_push(item_map)

        return output

    def current_workstep(self) -> AnalysisWorkstepTemplate:
        # noinspection PyTypeChecker
        return super().current_workstep()

    def _branch_current_workstep(self):
        current_workstep_template = self.current_workstep()
        new_workstep = AnalysisWorkstepTemplate(
            _common.new_placeholder_guid(), self, current_workstep_template.template, package=self.package)

        self.worksteps[new_workstep['ID']] = new_workstep
        self.definition['Current Workstep ID'] = new_workstep['ID']

        return new_workstep

    def copy(self, label):
        existing_worksheet = self._find_existing_worksheet_template(self.template.id, label)
        if existing_worksheet is not None:
            return existing_worksheet

        return AnalysisWorksheetTemplate(label, self.workbook, self.template, package=self.package, is_copy=True)


class TopicDocumentTemplate(WorksheetTemplate, TopicDocument):
    _document: ReportTemplate

    def __init__(self, label: str, topic, topic_document: TopicDocument, *, package: ItemList = None,
                 is_copy: bool = False):
        if not isinstance(topic_document, TopicDocument):
            raise SPyTypeError('topic_document parameter must be a TopicDocument')

        WorksheetTemplate.__init__(self, label, topic_document, package, is_copy)
        TopicDocument.__init__(self, topic, add_defaults=False)

        self._document = ReportTemplate(label, topic_document.document, self, package, is_copy=is_copy)

        # Clear definition for clarity, since it's not used in a template
        self._definition = dict()

    @property
    def template(self) -> TopicDocument:
        # noinspection PyTypeChecker
        return self._template

    @property
    def parameters(self):
        if self._parameters is None:
            return self.workbook.parameters
        return self._parameters

    @parameters.setter
    def parameters(self, val):
        super()._set_parameters(val)

    def _code_variable(self):
        return 'document'

    def code_dict(self):
        return self._code_dict_for_topic([self.document])

    @property
    def document(self) -> ReportTemplate:
        return self._document

    def push(self, session: Session, pushed_workbook_id, item_map, datasource_output, existing_worksheet_identifiers,
             include_inventory, label=None, status: Status = None):
        output = super().push(session, pushed_workbook_id,
                              OverrideItemMap(item_map, template_parameters=self.parameters), datasource_output,
                              existing_worksheet_identifiers, include_inventory, self.label, status=status)

        self._after_push(item_map)

        return output

    def copy(self, label):
        existing_worksheet = self._find_existing_worksheet_template(self.template.id, label)
        if existing_worksheet is not None:
            return existing_worksheet

        return TopicDocumentTemplate(label, self.workbook, self.template, package=self.package, is_copy=True)


class AnalysisWorkstepTemplate(ItemTemplate, AnalysisWorkstep):
    def __init__(self, label: str, worksheet, workstep: AnalysisWorkstep, *, package: ItemList = None,
                 is_copy: bool = False):
        if not isinstance(workstep, AnalysisWorkstep):
            raise SPyTypeError('workstep parameter must be a AnalysisWorkstep')

        ItemTemplate.__init__(self, label, workstep, package, is_copy)
        AnalysisWorkstep.__init__(self, worksheet)

        # Clear definition for clarity, since it's not used in a template
        self._definition = dict()

        # Copy the Data property to the wrapper class so that it can be overridden
        self['Data'] = copy.deepcopy(workstep['Data'])

    @property
    def template(self) -> AnalysisWorkstep:
        # noinspection PyTypeChecker
        return self._template

    @property
    def parameters(self):
        if self._parameters is None:
            return self.worksheet.parameters
        return self._parameters

    @parameters.setter
    def parameters(self, val):
        super()._set_parameters(val)

    def _code_variable(self):
        return 'workstep'

    def code_dict(self):
        return ItemTemplate._code_dict_for_analysis(self.template.worksheet.workbook, [self.template])

    def push_to_specific_worksheet(self, session: Session, pushed_workbook_id, pushed_worksheet_output,
                                   item_map: OverrideItemMap, include_inventory, *,
                                   existing_workstep_walk: Optional[ExistingWorkstepWalk] = None,
                                   no_workstep_message=None):
        output = super().push_to_specific_worksheet(
            session, pushed_workbook_id, pushed_worksheet_output,
            OverrideItemMap(item_map, template_parameters=self.parameters), include_inventory,
            existing_workstep_walk=existing_workstep_walk, no_workstep_message=no_workstep_message)

        item_map.override(self.template.id, item_map[self.id])

        self._after_push(item_map)

        return output

    def _find_existing_workstep_template(self, workstep_id, label):
        for workstep in self.worksheet.worksteps.values():
            if workstep.id == f'{workstep_id} {label}':
                return workstep

        return None

    def copy(self, label):
        existing_workstep = self._find_existing_workstep_template(self.template.id, label)
        if existing_workstep is not None:
            return existing_workstep

        return AnalysisWorkstepTemplate(label, self.worksheet, self.template, package=self.package, is_copy=True)


class AnnotationTemplate(Annotation):
    def __init__(self, label: str, annotation: Annotation, worksheet: Worksheet, annotation_type: str,
                 is_copy: bool = False):
        if not isinstance(annotation, Annotation):
            raise SPyTypeError('annotation parameter must be a Annotation')

        Annotation.__init__(self, worksheet, annotation_type)

        self.label = label
        self.template = annotation
        self.is_copy = is_copy

    @property
    def id(self):
        return f'{self.template.id} {self.label}'

    @property
    def html(self):
        return self.template.html

    @property
    def images(self):
        return self.template.images

    def resolve_content_code_key(self, content_id):
        return content_id

    @staticmethod
    def _walk_dict(_dict: dict, f: Callable):
        n = dict()
        for k, v in _dict.items():
            if isinstance(v, dict):
                n[k] = AnnotationTemplate._walk_dict(v, f)
            elif isinstance(v, list):
                n[k] = AnnotationTemplate._walk_list(v, f)
            else:
                n[k] = f(k, v)

        return n

    @staticmethod
    def _walk_list(_list: list, f: Callable):
        n = list()
        for i in _list:
            if isinstance(i, dict):
                n.append(AnnotationTemplate._walk_dict(i, f))
            elif isinstance(i, list):
                n.append(AnnotationTemplate._walk_list(i, f))
            else:
                n.append(i)

        return n

    def _push_specific(self, session: Session, item_map: OverrideItemMap, datasource_output, label, new_annotation,
                       existing_annotation, override_content_dict=None, status: Status = None):
        item_map.override(self.template.id, existing_annotation.id)

        html = super()._push_specific(session, item_map, datasource_output, label, new_annotation, existing_annotation,
                                      override_content_dict=override_content_dict, status=status)

        return html

    def _after_push(self, item_map):
        if not self.is_copy:
            # We also map to the non-templated ID in case there's a piece of content that has no template parameters
            item_map[self.template.id] = item_map[self.id]


class JournalTemplate(AnnotationTemplate, Journal):
    def __init__(self, label: str, journal: Journal, worksheet: AnalysisWorksheet, package: ItemList, *,
                 is_copy: bool = False):
        if not isinstance(journal, Journal):
            raise SPyTypeError('journal parameter must be a Journal')

        AnnotationTemplate.__init__(self, label, journal, worksheet, 'Journal', is_copy=is_copy)

    def _push_specific(self, session: Session, item_map: OverrideItemMap, datasource_output, label, new_annotation,
                       existing_annotation, override_content_dict=None, status: Status = None):
        parameters = item_map.parameters

        if parameters is None:
            parameters = self.worksheet.code_dict()

        html = super()._push_specific(session, item_map, datasource_output, label, new_annotation, existing_annotation)

        mustache = MustachioedAnnotation(html)
        new_html = mustache.render(parameters)

        self._after_push(item_map)

        return new_html


class ReportTemplate(AnnotationTemplate, Report):
    def __init__(self, label: str, report: Report, topic_document: TopicDocumentTemplate, package: ItemList, *,
                 is_copy: bool = False):
        if not isinstance(report, Report):
            raise SPyTypeError('report parameter must be a Report')

        AnnotationTemplate.__init__(self, label, report, topic_document, 'Report', is_copy=is_copy)

        def _to_template(_input_dict, _clazz):
            _dict = dict()
            for _id, _dr in _input_dict.items():
                _instance = _clazz(label, _dr, self, package=package, is_copy=is_copy)
                _dict[_id] = _instance

            return _dict

        self.date_ranges = _to_template(report.date_ranges, DateRangeTemplate)
        self.asset_selections = _to_template(report.asset_selections, AssetSelectionTemplate)
        self.content = _to_template(report.content, ContentTemplate)

        self.schedule = report.schedule
        self.data = report.data

    def resolve_content_code_key(self, content_id):
        content = self.content[content_id]
        workbook_templates = [w for w in self.worksheet.package if w.template.id == content['Workbook ID']]
        if len(workbook_templates) == 0:
            raise SPyValueError(f'Content with ID "{content.id}" references workbook ID "'
                                f'{content["Workbook ID"]}" but that workbook was not loaded at the same time as '
                                f'this topic.')

        workbook_template = workbook_templates[0]
        # noinspection PyUnresolvedReferences
        workbook = workbook_template.template
        worksheet = workbook.worksheets[content['Worksheet ID']]
        return ItemTemplate.code_key_string(content_id, Reference.EMBEDDED_CONTENT, worksheet.fqn)

    def _push_specific(self, session: Session, item_map: OverrideItemMap, datasource_output, label, new_annotation,
                       existing_annotation, override_content_dict=None, status: Status = None):
        content_dict = self.content.copy()

        def _workstep_to_content(_key, _val):
            _id, _type, _fqn = ItemTemplate.code_key_tuple(_key)
            if _id is None or _id not in content_dict:
                return _val

            if _val is None:
                return _id

            _content: ContentTemplate = content_dict[_id]
            _new_content = _content.copy(f'{self.label} {_val}')

            if isinstance(_val, AnalysisWorkstepTemplate):
                _new_content['Workbook ID'] = _val.worksheet.workbook.id
                _new_content['Worksheet ID'] = _val.worksheet.id
                _new_content['Workstep ID'] = _val.id
            elif isinstance(_val, AnalysisWorksheetTemplate):
                _new_content['Workbook ID'] = _val.workbook.id
                _new_content['Worksheet ID'] = _val.id
                _new_content['Workstep ID'] = _val.current_workstep().id
            else:
                raise SPyTypeError(f'Template parameter {_key} is of type {type(_val)} -- needs to '
                                   f'be AnalysisWorksheetTemplate or AnalysisWorkstepTemplate')

            content_dict[_new_content.id] = _new_content

            return _new_content

        parameters = item_map.parameters

        if parameters is None:
            parameters = self.worksheet.code_dict()

        parameters_with_content = AnnotationTemplate._walk_dict(parameters, _workstep_to_content)

        html = super()._push_specific(session, item_map, datasource_output, label, new_annotation, existing_annotation,
                                      content_dict, status)

        def _content_to_id(_key, _val):
            if not isinstance(_val, ContentTemplate):
                return _val

            return item_map[_val.id]

        parameters_with_content_ids = AnnotationTemplate._walk_dict(parameters_with_content, _content_to_id)

        mustache = MustachioedAnnotation(html, lambda content_id: self.resolve_content_code_key(content_id))

        new_html = mustache.render(parameters_with_content_ids)

        self._after_push(item_map)

        return new_html


class DateRangeTemplate(ItemTemplate, DateRange):
    def __init__(self, label: str, date_range: DateRange, report_template: ReportTemplate, package: ItemList, *,
                 is_copy: bool = False):
        if not isinstance(date_range, DateRange):
            raise SPyTypeError('date_range parameter must be a DateRange')

        ItemTemplate.__init__(self, label, date_range, package, is_copy)
        DateRange.__init__(self, date_range, report_template)

        # Clear definition for clarity, since it's not used in a template
        self._definition = dict()

    def push(self, session: Session, item_map: OverrideItemMap, existing_date_ranges: dict, status: Status):
        output = super().push(session, item_map, existing_date_ranges, status)

        item_map.override(self.template.id, item_map[self.id])

        self._after_push(item_map)

        return output


class AssetSelectionTemplate(ItemTemplate, AssetSelection):
    def __init__(self, label: str, asset_selection: AssetSelection, report_template: ReportTemplate,
                 *, package: ItemList = None, is_copy: bool = False):
        if not isinstance(asset_selection, AssetSelection):
            raise SPyTypeError('asset_selection parameter must be a AssetSelection')

        ItemTemplate.__init__(self, label, asset_selection, package, is_copy)
        AssetSelection.__init__(self, asset_selection, report_template)

        # Clear definition for clarity, since it's not used in a template
        self._definition = dict()

    def push(self, session: Session, item_map: OverrideItemMap, existing_asset_selections: dict, status: Status):
        output = super().push(session, item_map, existing_asset_selections, status)

        item_map.override(self.template.id, item_map[self.id])

        self._after_push(item_map)

        return output


class ContentTemplate(ItemTemplate, Content):
    template: Content

    def __init__(self, label: str, content: Content, report_template: ReportTemplate, package: ItemList, *,
                 is_copy: bool = False):
        if not isinstance(content, Content):
            raise SPyTypeError('content parameter must be a Content')

        ItemTemplate.__init__(self, label, content, package, is_copy)
        Content.__init__(self, content, report_template)

    def copy(self, label):
        return ContentTemplate(label, self.template, self.report, package=self.package, is_copy=True)

    def push(self, session: Session, item_map: OverrideItemMap, existing_contents: dict, status: Status):
        output = super().push(session, item_map, existing_contents, status)

        item_map.override(self.template.id, item_map[self.id])

        self._after_push(item_map)

        return output
